/*

This is the operating code for the FCW Model-2 robot by Foothills Community Workshop.

This code uses the NilRTOS real time OS. It has the following six threads running at the same time.
They are listed in priority order.

ReceiveRadioThread - Receive radio data.

ButtonThread - Catch the button presses used to change program modes.
  One press: Autonomous mode. Move around and avoid obstacles (default mode).
  Two presses: Dance mode. Bot becomes the leader and sends command to all bots to perform a little dance.
  Three presses: Calibrate Servos mode. Sets both of the servos to 90 degrees allowing them to be calibrated.

PingThread - Continuously ping the ultrasound module and calculate the distance to an object.

SendRadioThread - Use the radio to broadcast dance moves to all of the other bots.

ServoThread - Send commands to the servos.

TestThread - Test the buzzer, servos, ultrasound module and radio
*/

#include <nil.h>
#include <nilconf.h>
#include <nilcore.h>
#include <NilFIFO.h>
#include <NilRTOS.h>
#include <niltypes.h>
#include <SPI.h>

#include <NRF24.h>
#include <Ping.h>
#include <Servo360.h>

void Beep(int num);

typedef  struct {
  uint8_t DMove;
  uint16_t DNum;
} DanceMove;

union MoveData {
  DanceMove M;
  byte Buffer[sizeof(DanceMove)];  
};

//Servo direction parameters
const int DForward = 1;
const int DBackward = 2;
const int DRight = 3;
const int DLeft = 4;

//Radio data moves
const int MStart = 1;
const int MEnd = 2;
const int MForward = 3;
const int MBackward = 4;
const int MRight = 5;
const int MLeft = 6;
const int MStop = 7;
const int MBeep = 8;
const int MTest = 9;

//Arduino pins
const int Echo = 2;
const int NRFIrq = 3;
const int Trig = 4;
const int RServo = 5;
const int LServo = 6;
const int BeepPin = 7;
const int Button1 = 8;
const int CE = 9;
const int CSN = 10;

const int TooClose = 12;
const int DebounceTime = 200; //Button debounce milliseconds
const int BtnPushTimeout = 1000; //Maximum time, in milliseconds, between button presses in a button command
const int BeepLen = 100; //Length of a beep in milliseconds

//This defines the dance
const DanceMove DMoves[] = {{MStop, 1000}, {MForward,500}, {MBackward,500}, {MRight,1500}, {MBeep,1}, {MLeft,1500}, {MBeep,1}};
const int NumMoves = sizeof(DMoves) / sizeof(DanceMove);

//Program modes
const int Autonomous = 1;
const int Dance = 2;
const int CalibrateServo = 3;
const int TestMode = 4;

//Current program mode
uint8_t ProgramMode = Autonomous;

//Have we received the dance start move?
bool Dancing = false;

//Are we in Leader mode. The Leader transmits dance moves to the other robots.
bool Leader = false;

//Current obstacle distance
long inches;

//Servo objects
Servo Right;
Servo Left;

//Storage for the received dance move
MoveData Move = {0,0};

//nRF24L01 radio variable
NRF24 radio;

//Semaphore used by ButtonThread
SEMAPHORE_DECL(SemBtnPress, 0);

//Semaphore used by RadioThread
SEMAPHORE_DECL(SemRadio, 0);

//NilRTOS thread stacks
NIL_WORKING_AREA(waReceiveRadioThread, 128);
NIL_WORKING_AREA(waButtonThread, 128);
NIL_WORKING_AREA(waPingThread, 128);
NIL_WORKING_AREA(waSendRadioThread, 128);
NIL_WORKING_AREA(waServoThread, 128);
NIL_WORKING_AREA(waTestThread, 128);

//Declarations of thread functions
NIL_THREAD(ReceiveRadioThread, arg);
NIL_THREAD(ButtonThread, arg);
NIL_THREAD(PingThread, arg);
NIL_THREAD(SendRadioThread, arg);
NIL_THREAD(ServoThread, arg);
NIL_THREAD(TestThread, arg);

//CreateNilRTOS threads. They must be created in priority order.
NIL_THREADS_TABLE_BEGIN()
NIL_THREADS_TABLE_ENTRY("ReceiveRadioThread", ReceiveRadioThread, NULL, waReceiveRadioThread, sizeof(waReceiveRadioThread))
NIL_THREADS_TABLE_ENTRY("SendRadioThread", SendRadioThread, NULL, waSendRadioThread, sizeof(waSendRadioThread))
NIL_THREADS_TABLE_ENTRY("ServoThread", ServoThread, NULL, waServoThread, sizeof(waServoThread))
NIL_THREADS_TABLE_ENTRY("ButtonThread", ButtonThread, NULL, waButtonThread, sizeof(waButtonThread))
NIL_THREADS_TABLE_ENTRY("PingThread", PingThread, NULL, waPingThread, sizeof(waPingThread))
NIL_THREADS_TABLE_ENTRY("TestThread", TestThread, NULL, waTestThread, sizeof(waTestThread))
NIL_THREADS_TABLE_END()


//Consume a dance move
bool Consume(union MoveData *DM){
  bool ret = false;
  
  //Make the following code atomic
  nilSysLock();
  
  if (Move.M.DMove != 0){
    *DM = Move;

    Move.M = (DanceMove) {0,0};
 
    ret = true;
  }
  
  //Finished with atomic code
  nilSysUnlock();
  
  return ret;
};

//Produce a dance move
bool Produce(union MoveData DM){
  bool ret = false;
  
  nilSysLock();
  
  if (Move.M.DMove == 0){
    Move = DM;

    ret = true;
  }
  
  nilSysUnlock();
  
  return ret;
};

//Interrupt 0 ISR, Button press interrupt
ISR (PCINT0_vect) {
  static unsigned long Time = 0;
  unsigned long Now = millis();
  
  NIL_IRQ_PROLOGUE();
  
  if (Now - Time >= DebounceTime) {
    Time = Now;

    if (!digitalRead(Button1))
      nilSemSignal(&SemBtnPress);
  }
  
  NIL_IRQ_EPILOGUE();
}

//Radio interrupt
void RadioInterrupt() {
  NIL_IRQ_PROLOGUE();
 
  nilSemSignalI(&SemRadio);
  
  NIL_IRQ_EPILOGUE();
}

//Button thread
NIL_THREAD(ButtonThread, arg) {
  int BtnPresses = 0;
  
  while(TRUE) {
    //wait for first button press
    nilSemWaitTimeout(&SemBtnPress,TIME_INFINITE);

    BtnPresses = 1;

    nilThdSleepMilliseconds(150);

    while(TRUE){
      //wait for all button presses
      if (nilSemWaitTimeout(&SemBtnPress,BtnPushTimeout) == NIL_MSG_TMO){
        if (ProgramMode != TestMode) {
          //No more button presses
          for (int i = 0; i < BtnPresses; i++) {
            nilThdSleepMilliseconds(150);
            Beep(1);          
          }
        
          switch (BtnPresses) {
            case Autonomous:
              ProgramMode = Autonomous;
              
              if (Leader)
                LeaderReset();
              
              break;
            case Dance:
              ProgramMode = Autonomous;
              Leader = true;
              
              break;
            case CalibrateServo:
              ProgramMode = CalibrateServo;
  
              ServoCalibrate();
              break;
          }
        }
        
        nilThdSleepMilliseconds(100);
        break;        
      }
      else {
        BtnPresses++;
      }
    }
  }
}

//Robot is no longer a Leader
void LeaderReset(){
  Leader = false;
  Dancing = false;
  
  radio.listenToAddress(0xD2);
}

NIL_THREAD(ReceiveRadioThread, arg) {
  union MoveData MyData;
    
  while(true) {
    nilSemWait(&SemRadio);
    
    if (radio.available()){
      radio.read(MyData.Buffer, sizeof(MyData.Buffer));
      
      if (ProgramMode == Autonomous && !Leader) {
        if (MyData.M.DMove == MStart)
          Dancing = true;
        else {
          while(!Produce(MyData)) {
            nilThdSleepMilliseconds(50);
          }
        }  
      }
      else if (ProgramMode == TestMode && MyData.M.DMove == MTest){
        BeepVarLen(BeepLen / 2);
        BeepVarLen(BeepLen / 2);
      }
    }
  }
}

//Radio thread
NIL_THREAD(SendRadioThread, arg) {
  union MoveData MyData;
  bool DanceEnded = false;

  while(true) {
    if (ProgramMode == Autonomous) {
      if (Leader){
        if (!Dancing) {
          Dancing = true;
          
          MyData.M = (DanceMove){MStart, 0};
          
          radio.broadcast(MyData.Buffer, sizeof(MyData.Buffer));
          DanceEnded = false;
        }
        else{
          if (!DanceEnded) {
            for (int i = 0; i <= NumMoves; i++){  
              if (i == NumMoves) {
                MyData.M = (DanceMove) {MEnd,0};
                DanceEnded = true;
              }
              else {
                MyData.M = (DanceMove){DMoves[i].DMove,DMoves[i].DNum};
              }    
                  
              while(!Produce(MyData))
                nilThdSleepMilliseconds(50);
              
              radio.broadcast(MyData.Buffer, sizeof(MyData.Buffer));
            }
          }
        }
      }
    }
    
    nilThdSleepMilliseconds(50);
  }
}

//Ping sensor thread
NIL_THREAD(PingThread, arg) {
  while(true) {
    if (ProgramMode == Autonomous || ProgramMode == TestMode)
      inches = DoPing(Trig, Echo);
      
    nilThdSleepMilliseconds(90);
  }
}

//Servo turn function
void Turn(int RL){
  if (RL == DRight) {
    Right.write(180);
    Left.write(180);
  }
  else if (RL == DLeft) {
    Right.write(0);
    Left.write(0);
  }
}

//Servo turn randomly right or left function
void TurnRandom() {
  if (random(1,10001) > 5000) {
    Turn(DRight);
  }
  else {
    Turn(DLeft);
  }
}

//Servo stop function
void Stop() {
  Right.write(90);
  Left.write(90);
}

//Servo move forward or backward function
void Run(int FB) {
  if (FB == DForward) {
    Right.write(0);
    Left.write(180);
  }
  else if (FB == DBackward){
    Right.write(180);
    Left.write(0);
  }
}

//Servo thread
NIL_THREAD(ServoThread, arg) {
  bool Wait;
  union MoveData MyData;
  
  while (true) {
    if (ProgramMode == Autonomous) {
      if (!Dancing) {
        if (inches < TooClose)
        {
          TurnRandom();
          Beep(1);
        }
        else
        {
          //Run straight
          Run(DForward);
        }
      }
      else {
        Wait = true;      

        while(!Consume(&MyData))
          nilThdSleepMilliseconds(50);

        switch (MyData.M.DMove) {
          case MForward:
            Run(DForward);
            break;
          case MBackward:
            Run(DBackward);
            break;
          case MRight:
            Turn(DRight);
            break;
          case MLeft:
            Turn(DLeft);
            break;
          case MStop:
            Stop();
            break;
          case MEnd:
            LeaderReset();
            break;
          case MBeep:
            Beep(MyData.M.DNum);
            Wait = false;
            break;
          default:
            Wait = false;
            break;
        }
        if (Wait && MyData.M.DNum > 0)
          nilThdSleepMilliseconds(MyData.M.DNum);
      }
    }
    
    nilThdSleepMilliseconds(60);
  }
} 

//Ping Test thread
NIL_THREAD(TestThread, arg) {
  union MoveData MyData;

  while(true) {
    if (ProgramMode == TestMode) {
      //Beep twice
      Beep(2);
      
      //Run right servo forward then back
      Right.attach(RServo);  // attaches the right servo to the servo object 
      Right.write(0);
      nilThdSleepMilliseconds(400);
      Right.write(180);
      nilThdSleepMilliseconds(400);
      Right.detach();
    
      //Run left servo forward then back
      Left.attach(LServo);  // attaches the left servo to the servo object   
      Left.write(180);
      nilThdSleepMilliseconds(400);

      Left.write(0);
      nilThdSleepMilliseconds(400);
      Left.detach();

      while(true) {
        if (inches < 6) {
          Beep(1);
        }
  
        if (!digitalRead(Button1)) {
          Beep(2);
          break;
        }
        
        nilThdSleepMilliseconds(100);
      };
      
      MyData.M = (DanceMove){MTest, 0};
      
      while(true) {
        if (!digitalRead(Button1)) {
          radio.broadcast(MyData.Buffer, sizeof(MyData.Buffer));
          nilThdSleepMilliseconds(50);
          radio.listenToAddress(0xD2);
        }

        nilThdSleepMilliseconds(500);
      }
    }
    
    ProgramMode = Autonomous;
  };
}

//Stop the servos so they can be calibrated
void ServoCalibrate(){
  Stop();
}

void setup() {
  randomSeed(analogRead(0));
  pinMode(BeepPin,OUTPUT);
  digitalWrite(BeepPin, LOW);

  pinMode(Button1,INPUT_PULLUP);

  //
  //Configure rf radio on SPI bus 
  //
  radio.begin(CE, CSN);
  radio.setAddress(0xD2);
  radio.listenToAddress(0xD2);

  if (!digitalRead(Button1))
    ProgramMode = TestMode;

  //Start servos    
  Right.attach(RServo);  // attaches the right servo to the servo object 
  Left.attach(LServo);  // attaches the left servo to the servo object   

  attachInterrupt(1, &RadioInterrupt, FALLING);
  
  //Setup pin change interrupt for tactile button
  noInterrupts();
  //Setup for Button pin change interrupt
  PCMSK0 |= bit (PCINT0);  // want pin D8
  PCIFR  |= bit (PCIF0);   // clear any outstanding interrupts
  PCICR  |= bit (PCIE0);   // enable pin change interrupts for D8  
  interrupts();

  //Start NilRTOS.  
  nilSysBegin();
}

//Beep routine. num parameter is number of beeps
void Beep(int num)
{
  for (int i = 0; i < num; i++)
    BeepVarLen(BeepLen);
}

void BeepVarLen(int Len){
  digitalWrite(BeepPin,HIGH);
  nilThdSleepMilliseconds(Len);
  digitalWrite(BeepPin,LOW);    
  nilThdSleepMilliseconds(Len);
}

void loop() {
  // Not used.
}

